/*
 * linux/drivers/video/ep93xxfb.c -- Frame buffer device for the EP93xx
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License. See the file COPYING in the main directory of this archive for
 *  more details.
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/fb.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/dma-mapping.h>
#include <linux/reboot.h>
#include <video/ep93xxfb.h>
#include <asm/arch/cx25871.h>

#define devicename				"ep93xxfb"

struct ep93xxfb_videotimings
{
	const char *name;
	void (*configure)(struct fb_info *info);
	void (*on)(struct fb_info *info);
	void (*off)(struct fb_info *info);

	unsigned long hres;
	unsigned long hfp;
	unsigned long hsync;
	unsigned long hbp;

	unsigned long vres;
	unsigned long vfp;
	unsigned long vsync;
	unsigned long vbp;

	unsigned long refresh;
	unsigned long width;
	unsigned long height;

	unsigned long clk_src:1;
	unsigned long clk_edge:1;
	unsigned long pol_blank:1;
	unsigned long pol_hsync:1;
	unsigned long pol_vsync:1;
};

#define POL_HIGH				1
#define POL_LOW					0
#define EDGE_RISING				1
#define EDGE_FALLING				0
#define CLK_INTERNAL				1
#define CLK_EXTERNAL				0

#define GPIO_DR					GPIO_PGDR
#define GPIO_DDR				GPIO_PGDDR
#define GPIO_CLK				(1 << 0)
#define GPIO_DAT				(1 << 1)
#define DELAY_USEC				2
#define READ_TIMEOUT				100


#define EE_DELAY_USEC			2
#define EE_READ_TIMEOUT			100
#define CX25871_DEV_ADDRESS		0x88


#define GPIOG_EEDAT 2
#define GPIOG_EECLK 1

static int
write_reg(unsigned char ucRegAddr, unsigned char ucRegValue)
{
  	unsigned long uiVal, uiDDR;
	unsigned char ucData, ucIdx, ucBit;
	unsigned long ulTimeout;

	uiVal = inl(GPIO_PGDR);
	uiDDR = inl(GPIO_PGDDR);

	uiVal |= (GPIOG_EEDAT | GPIOG_EECLK);
	outl( uiVal, GPIO_PGDR );

	udelay( EE_DELAY_USEC );

	uiDDR |= (GPIOG_EEDAT | GPIOG_EECLK);
	outl( uiDDR, GPIO_PGDDR );

	udelay( EE_DELAY_USEC );

	uiVal &= ~GPIOG_EEDAT;
	outl( uiVal, GPIO_PGDR );

	udelay( EE_DELAY_USEC );

	uiVal &= ~GPIOG_EECLK;
	outl( uiVal, GPIO_PGDR );

	udelay( EE_DELAY_USEC );

	for (ucIdx = 0; ucIdx < 3; ucIdx++) {
		if (ucIdx == 0)
			ucData = (unsigned char)CX25871_DEV_ADDRESS;
		else if (ucIdx == 1)
			ucData = ucRegAddr;
		else
			ucData = ucRegValue;

		for (ucBit = 0; ucBit < 8; ucBit++) {
			if (ucData & 0x80)
				uiVal |= GPIOG_EEDAT;
			else
				uiVal &= ~GPIOG_EEDAT;
			outl( uiVal, GPIO_PGDR );

			udelay( EE_DELAY_USEC );

			outl( (uiVal | GPIOG_EECLK), GPIO_PGDR );

			udelay( EE_DELAY_USEC );

			outl( uiVal, GPIO_PGDR );

			udelay( EE_DELAY_USEC );

			ucData <<= 1;
		}

		uiDDR &= ~GPIOG_EEDAT;
		outl( uiDDR, GPIO_PGDDR );

		udelay( EE_DELAY_USEC );

		outl( (uiVal | GPIOG_EECLK), GPIO_PGDR );

		udelay( EE_DELAY_USEC );

		ulTimeout = 0;
		while ( inl(GPIO_PGDR) & GPIOG_EEDAT ) {
			udelay( EE_DELAY_USEC );
			ulTimeout++;
			if (ulTimeout > EE_READ_TIMEOUT )
				return 1;
		}

		outl( uiVal, GPIO_PGDR );

		udelay( EE_DELAY_USEC );

		uiDDR |= GPIOG_EEDAT;
		outl( uiDDR, GPIO_PGDDR );
		udelay( EE_DELAY_USEC );
	}

	uiVal &= ~GPIOG_EEDAT;
	outl( uiVal, GPIO_PGDR );

	udelay( EE_DELAY_USEC );

	uiVal |= GPIOG_EECLK;
	outl( uiVal, GPIO_PGDR );

	udelay( EE_DELAY_USEC );

	uiVal |= GPIOG_EEDAT;
	outl( uiVal, GPIO_PGDR );

	udelay( EE_DELAY_USEC );


	return 0;
}

static void
philips_lb064v02_on(struct fb_info *info)
{
	outl(inl(GPIO_PADDR) | 2, GPIO_PADDR);
	outl(inl(GPIO_PADR) | 2, GPIO_PADR);
}

static void
philips_lb064v02_off(struct fb_info *info)
{
	outl(inl(GPIO_PADR) & ~2, GPIO_PADR);
}

static int
conexant_cx25871_reboot(struct notifier_block *this, unsigned long event,
			void *x)
{
	switch (event) {
	case SYS_HALT:
	case SYS_POWER_OFF:
	case SYS_RESTART:
		write_reg(0xba, 0x80); //0x88
	default:
		return NOTIFY_DONE;
	}
}

static struct notifier_block conexant_cx25871_notifier = {
	conexant_cx25871_reboot,
	NULL,
	5
};

static void
conexant_cx25871_configure_ntsc(struct fb_info *info)
{
	write_reg(0xb8, 0x00);
       	write_reg(0xBA, CX25871_REGxBA_SLAVER | CX25871_REGxBA_DACOFF);     
	write_reg(0xC6, (CX25871_REGxC6_INMODE_MASK & 0x3));
	write_reg(0xC4, CX25871_REGxC4_EN_OUT);
	write_reg(0x32, 0);
	write_reg(0xBA, CX25871_REGxBA_SLAVER );

	register_reboot_notifier(&conexant_cx25871_notifier);
}

static void
conexant_cx25871_configure_pal(struct fb_info *info)
{
	write_reg(0xb8, 0x01);
       	write_reg(0xBA, CX25871_REGxBA_SLAVER | CX25871_REGxBA_DACOFF);     
	write_reg(0xC6, (CX25871_REGxC6_INMODE_MASK & 0x3));
	write_reg(0xC4, CX25871_REGxC4_EN_OUT);
	write_reg(0x32, 0);
	write_reg(0xBA, CX25871_REGxBA_SLAVER );

	register_reboot_notifier(&conexant_cx25871_notifier);
}



static void
conexant_cx25871_overscan_configure(struct fb_info *info)
{
	write_reg(0xb8, 0x00);
	write_reg(0xba, 0x30);
	write_reg(0x34, 0x9b);
	write_reg(0x36, 0xc0);
	write_reg(0xc8, 0x92);
	write_reg(0xca, 0xd3);
	write_reg(0xcc, 0xd3);
	write_reg(0xd8, 0x60);
	write_reg(0x38, 0x00);
	write_reg(0x76, 0x10);
	write_reg(0x78, 0x80);
	write_reg(0x7a, 0x72);
	write_reg(0x7c, 0x82);
	write_reg(0x7e, 0x42);
	write_reg(0x80, 0xF5);
	write_reg(0x82, 0x13);
	write_reg(0x84, 0xF2);
	write_reg(0x86, 0x26);
	write_reg(0x88, 0x00);
	write_reg(0x8a, 0x08);
	write_reg(0x8c, 0x77);
	write_reg(0x8e, 0x03);
	write_reg(0x90, 0x0D);
	write_reg(0x92, 0x24);
	write_reg(0x94, 0xE0);
	write_reg(0x96, 0x06);
	write_reg(0x98, 0x00);
	write_reg(0x9a, 0x10);
	write_reg(0x9c, 0x68);
	write_reg(0x9e, 0xDA);
	write_reg(0xa0, 0x0A);
	write_reg(0xa2, 0x0A);
	write_reg(0xa4, 0xE5);
	write_reg(0xa6, 0x77);
	write_reg(0xa8, 0xC2);
	write_reg(0xaa, 0x8A);
	write_reg(0xac, 0x9A);
	write_reg(0xae, 0x12);
	write_reg(0xb0, 0x99);
	write_reg(0xb2, 0x86);
	write_reg(0xb4, 0x25);
	write_reg(0xb6, 0x00);
	write_reg(0xbc, 0x00);
	write_reg(0xbe, 0x00);
	write_reg(0xc0, 0x00);
	write_reg(0xc2, 0x00);
	mdelay(1);
	write_reg(0x6c, 0xc4);
	write_reg(0xc6, 0x03);
	write_reg(0xc4, 0x01);
	write_reg(0x32, 0x00);
	register_reboot_notifier(&conexant_cx25871_notifier);
}

static void
conexant_cx25871_on(struct fb_info *info)
{
	write_reg(0xba, 0x20);
}

static void
conexant_cx25871_off(struct fb_info *info)
{
	write_reg(0xba, 0x30);
}

typedef enum
{
	CRT_VGA,
	CRT_VGA75,
	CRT_SVGA,
	CRT_SVGA75,
	CRT_XGA,
	CRT_XGA75,
	CRT_XGAPLUS,
	CRT_XGAPLUS75,
	CRT_SXGA,
	PHILIPS_LB064V02,
	CX25871_NTSC,
	CX25871_OVERSCAN,
	CX25871_PAL,
} DisplayType;

static struct ep93xxfb_videotimings modedb[] = {
	[CRT_VGA] = {
		.name =		"CRT-640x480",
		.hres =		640,
		.hfp =		16,
		.hsync =	96,
		.hbp =		48,
		.vres =		480,
		.vfp =		10,
		.vsync =	2,
		.vbp =		33,
		.refresh =	60,
		.width =	640,
		.height =	480,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_LOW,
		.pol_vsync =	POL_LOW,
	},
	[CRT_VGA75] = {
		.name =		"CRT-640x480@75",
		.hres =		640,
		.hfp =		16,
		.hsync =	64,
		.hbp =		120,
		.vres =		480,
		.vfp =		1,
		.vsync =	3,
		.vbp =		16,
		.refresh =	75,
		.width =	640,
		.height =	480,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_LOW,
		.pol_vsync =	POL_LOW,
	},
	[CRT_SVGA] = {
		.name =		"CRT-800x600",
		.hres =		800,
		.hfp =		40,
		.hsync =	128,
		.hbp =		88,
		.vres =		600,
		.vfp =		1,
		.vsync =	4,
		.vbp =		23,
		.refresh =	60,
		.width =	800,
		.height =	600,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_HIGH,
		.pol_vsync =	POL_HIGH,
	},
	[CRT_SVGA75] = {
		.name =		"CRT-800x600@75",
		.hres =		800,
		.hfp =		16,
		.hsync =	80,
		.hbp =		160,
		.vres =		600,
		.vfp =		1,
		.vsync =	3,
		.vbp =		21,
		.refresh =	75,
		.width =	800,
		.height =	600,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_HIGH,
		.pol_vsync =	POL_HIGH,
	},
	[CRT_XGA] = {
		.name =		"CRT-1024x768",
		.hres =		1024,
		.hfp =		24,
		.hsync =	136,
		.hbp =		160,
		.vres =		768,
		.vfp =		3,
		.vsync =	6,
		.vbp =		29,
		.refresh =	60,
		.width =	1024,
		.height =	768,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_LOW,
		.pol_vsync =	POL_LOW,
	},
	[CRT_XGA75] = {
		.name =		"CRT-1024x768@75",
		.hres =		1024,
		.hfp =		16,
		.hsync =	96,
		.hbp =		176,
		.vres =		768,
		.vfp =		1,
		.vsync =	3,
		.vbp =		28,
		.refresh =	75,
		.width =	1024,
		.height =	768,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_HIGH,
		.pol_vsync =	POL_HIGH,
	},
	[CRT_XGAPLUS] = {
		.name =		"CRT-1152x864",
		.hres =		1152,
		.hfp =		32,
		.hsync =	96,
		.hbp =		192,
		.vres =		864,
		.vfp =		1,
		.vsync =	3,
		.vbp =		37,
		.refresh =	60,
		.width =	1152,
		.height =	864,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_HIGH,
		.pol_vsync =	POL_HIGH,
	},
	[CRT_XGAPLUS75] = {
		.name =		"CRT-1152x864@75",
		.hres =		1152,
		.hfp =		51,//64,
		.hsync =	102,//128,
		.hbp =		205,//256,
		.vres =		864,
		.vfp =		1,
		.vsync =	2,//3,
		.vbp =		16,//32,
		.refresh =	75,
		.width =	1152,
		.height =	864,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_HIGH,
		.pol_vsync =	POL_HIGH,
	},
	[CRT_SXGA] = {
		.name =		"CRT-1280x1024",
		.hres =		1280,
		.hfp =		37,//48,
		.hsync =	86,//112,
		.hbp =		191,//248,
		.vres =		1024,
		.vfp =		1,
		.vsync =	2,//3
		.vbp =		19,//38,
		.refresh =	60,
		.width =	1280,
		.height =	1024,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_RISING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_HIGH,
		.pol_vsync =	POL_HIGH,
	},
	[PHILIPS_LB064V02] = {
		.name =		"Philips-LB064V02",
		.on =		philips_lb064v02_on,
		.off =		philips_lb064v02_off,
		.hres =		640,
		.hfp =		24,
		.hsync =	96,
		.hbp =		40,
		.vres =		480,
		.vfp =		10,
		.vsync =	2,
		.vbp =		33,
		.refresh =	60,
		.width =	132,
		.height =	100,
		.clk_src =	CLK_INTERNAL,
		.clk_edge =	EDGE_FALLING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_LOW,
		.pol_vsync =	POL_LOW,
	},
	[CX25871_NTSC] = {
		.name =		"Conexant-CX25871-NTSC",
		.configure =	conexant_cx25871_configure_ntsc,
		.on =		conexant_cx25871_on,
		.off =		conexant_cx25871_off,
		.hres =		640,
		.hfp =		16,
		.hsync =	4,
		.hbp =		124,
		.vres =		480,
		.vfp =		45,
		.vsync =	2,
		.vbp =		73,
		.refresh =	35466,
		.width =	300,
		.height =	230,
		.clk_src =	CLK_EXTERNAL,
		.clk_edge =	EDGE_FALLING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_LOW,
		.pol_vsync =	POL_LOW,
	},
	[CX25871_OVERSCAN] = {
		.name =		"Conexant-CX25871-overscan",
		.configure =	conexant_cx25871_overscan_configure,
		.on =		conexant_cx25871_on,
		.off =		conexant_cx25871_off,
		.hres =		640,
		.hfp =		15,
		.hsync =	4,
		.hbp =		117,
		.vres =		480,
		.vfp =		9,
		.vsync =	2,
		.vbp =		34,
		.refresh =	40950,
		.width =	300,
		.height =	230,
		.clk_src =	CLK_EXTERNAL,
		.clk_edge =	EDGE_FALLING,
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_LOW,
		.pol_vsync =	POL_LOW,
	},
	[CX25871_PAL] = {
		.name =		"Conexant-CX25871-PAL",
		.configure =	conexant_cx25871_configure_pal,
		.on =		conexant_cx25871_on,
		.off =		conexant_cx25871_off,
		.hres =		640,
		.hfp =		50, //50
		.hsync =	4,//4 
		.hbp =		250, //250
		.vres =		480,
		.vfp =		50,
		.vsync =	2,
		.vbp =		93,
		.refresh =	35466,
		.width =	320, //300
		.height =	240, //230
		.clk_src =	CLK_EXTERNAL,
		.clk_edge =	EDGE_FALLING, 
		.pol_blank =	POL_LOW,
		.pol_hsync =	POL_LOW,
		.pol_vsync =	POL_LOW,
	},
};

#if defined(CONFIG_FB_EP93XX_CRT)
#if defined (CONFIG_FB_EP93XX_CRT_VGA)
#define DEFAULT_MODE CRT_VGA
#elif defined (CONFIG_FB_EP93XX_CRT_VGA75)
#define DEFAULT_MODE CRT_VGA75
#elif defined (CONFIG_FB_EP93XX_CRT_SVGA)
#define DEFAULT_MODE CRT_SVGA
#elif defined (CONFIG_FB_EP93XX_CRT_SVGA75)
#define DEFAULT_MODE CRT_SVGA75
#elif defined (CONFIG_FB_EP93XX_CRT_XGA)
#define DEFAULT_MODE CRT_XGA
#elif defined (CONFIG_FB_EP93XX_CRT_XGA75)
#define DEFAULT_MODE CRT_XGA75
#elif defined (CONFIG_FB_EP93XX_CRT_XGAPLUS)
#define DEFAULT_MODE CRT_XGAPLUS
#elif defined (CONFIG_FB_EP93XX_CRT_XGAPLUS75)
#define DEFAULT_MODE CRT_XGAPLUS75
#elif defined (CONFIG_FB_EP93XX_CRT_SXGA)
#define DEFAULT_MODE CRT_SXGA
#else
#error What Display Setting was that!!!
#endif
#elif defined(CONFIG_FB_EP93XX_LCD)
#define DEFAULT_MODE PHILIPS_LB064V02
#elif defined(CONFIG_FB_EP93XX_PAL)
#define DEFAULT_MODE CX25871_PAL
#elif defined(CONFIG_FB_EP93XX_NTSC)
#ifdef CONFIG_FB_EP93XX_NTSC_OVERSCAN
#define DEFAULT_MODE CX25871_OVERSCAN
#else
#define DEFAULT_MODE CX25871_NTSC
#endif
#else
#define DEFAULT_MODE CRT_VGA
#endif
#if defined(CONFIG_FB_EP93XX_8BPP)
#define DEFAULT_BPP  8
#elif defined(CONFIG_FB_EP93XX_16BPP)
#define DEFAULT_BPP 16
#elif defined(CONFIG_FB_EP93XX_24BPP)
#define DEFAULT_BPP 24
#elif defined(CONFIG_FB_EP93XX_32BPP)
#define DEFAULT_BPP 32
#else
#define DEFAULT_BPP 16
#endif

static unsigned int pseudo_palette[256];

#ifdef MODULE
static char *modestr = 0;
#endif
static int mode = DEFAULT_MODE;
static int bpp = DEFAULT_BPP;

#define ep93xxfb_outl(value, reg)				\
	{							\
		outl(RASTER_SWLOCK_VALUE, RASTER_SWLOCK);	\
		outl(value, reg);				\
	}

static inline void
ep93xxfb_palette_write(u_int regno, u_int red, u_int green,
		       u_int blue, u_int trans)
{
    unsigned int cont, i, pal;

    pal = ((red & 0xFF00) << 8) | (green & 0xFF00) | ((blue & 0xFF00) >> 8);

    pseudo_palette[regno] = pal;

    outl( pal, (COLOR_LUT+(regno<<2)) );

    cont = inl(LUTCONT);

    if ((cont&LUTCONT_STAT && cont&LUTCONT_RAM1) ||
	(!(cont&LUTCONT_STAT) && !(cont&LUTCONT_RAM1)))
    {
       	for (i=0; i< 256; i++)  // Update inactive LUT
	{
	    outl( pseudo_palette[i], (COLOR_LUT+(i<<2)) );
	}
	// Switch active LUTs next frame
	outl( cont ^ LUTCONT_RAM1, LUTCONT );
    }
}

static int
ep93xxfb_setcolreg(unsigned regno, unsigned red, unsigned green, unsigned blue,
		   unsigned transp, struct fb_info *info)
{

#define CNVT_TOHW(val,width) ((((val)<<(width))+0x7FFF-(val))>>16)

	switch (info->fix.visual) {
	case FB_VISUAL_PSEUDOCOLOR:
		ep93xxfb_palette_write(regno, red, green, blue, transp);
		break;
	case FB_VISUAL_TRUECOLOR:
		if (regno >= 16)
			return 1;

		red = CNVT_TOHW(red, info->var.red.length);
		green = CNVT_TOHW(green, info->var.green.length);
		blue = CNVT_TOHW(blue, info->var.blue.length);
		transp = CNVT_TOHW(transp, info->var.transp.length);
		((u32 *) (info->pseudo_palette))[regno] =
		  (red << info->var.red.offset) |
		  (green << info->var.green.offset) |
		  (blue << info->var.blue.offset) |
		  (transp << info->var.transp.offset);
		break;
	case FB_VISUAL_DIRECTCOLOR:
		red = CNVT_TOHW(red, 8);
		green = CNVT_TOHW(green, 8);
		blue = CNVT_TOHW(blue, 8);
		transp = CNVT_TOHW(transp, 8);
		break;
	}
#undef CNVT_TOHW

	return 0;
}

static int
ep93xxfb_blank(int blank_mode, struct fb_info *info)
{
	unsigned long attribs;

	attribs = inl(VIDEOATTRIBS) & ~VIDEOATTRIBS_SDSEL_MASK;
#ifdef CONFIG_EP93XX_SDCS0
	attribs |= 0 << VIDEOATTRIBS_SDSEL_SHIFT;
#endif
#ifdef CONFIG_EP93XX_SDCS1
	attribs |= 1 << VIDEOATTRIBS_SDSEL_SHIFT;
#endif
#ifdef CONFIG_EP93XX_SDCS2
	attribs |= 2 << VIDEOATTRIBS_SDSEL_SHIFT;
#endif
#ifdef CONFIG_EP93XX_SDCS3
	attribs |= 3 << VIDEOATTRIBS_SDSEL_SHIFT;
#endif

	if (blank_mode) {
		if (modedb[mode].off)
			(modedb[mode].off)(info);
		ep93xxfb_outl(attribs & ~(VIDEOATTRIBS_DATAEN |
					  VIDEOATTRIBS_SYNCEN |
					  VIDEOATTRIBS_PCLKEN |
					  VIDEOATTRIBS_EN),
		     VIDEOATTRIBS);
	} else {
		if (modedb[mode].clk_src == CLK_INTERNAL)
			attribs |= VIDEOATTRIBS_PCLKEN;
		ep93xxfb_outl(attribs | VIDEOATTRIBS_DATAEN
			      | VIDEOATTRIBS_SYNCEN
			      | VIDEOATTRIBS_EN,
			      VIDEOATTRIBS);
		if (modedb[mode].on)
			(modedb[mode].on)(info);
	}

	return 0;
}


/*
 * ep93xxfb_set_par():
 *
 */
static int
ep93xxfb_set_par(struct fb_info *info)
{
	
	switch (info->var.bits_per_pixel) {
	case 8:
		info->fix.visual = FB_VISUAL_PSEUDOCOLOR;
		break;

	case 16:
	case 24:
	case 32:
		info->fix.visual = FB_VISUAL_TRUECOLOR;
		break;

	default:
		return -EINVAL;
	}

	info->fix.line_length = (info->var.xres * info->var.bits_per_pixel) / 8;

	return 0;
}

static struct fb_ops ep93xxfb_ops = {
	.owner		= THIS_MODULE,
	.fb_setcolreg	= ep93xxfb_setcolreg,
	.fb_set_par     = ep93xxfb_set_par,
	.fb_blank	= ep93xxfb_blank,
	.fb_fillrect	= cfb_fillrect,
	.fb_copyarea	= cfb_copyarea,
	.fb_imageblit	= cfb_imageblit,
	.fb_cursor	= soft_cursor,
};

static unsigned long
ep93xx_get_pll_frequency(unsigned long pll)
{
	unsigned long fb1, fb2, ipd, ps, freq;

	if (pll == 1)
		pll = inl(SYSCON_CLKSET1);
	else if (pll == 2)
		pll = inl(SYSCON_CLKSET2);
	else
		return 0;

	ps = (pll & SYSCON_CLKSET1_PLL1_PS_MASK) >>
		SYSCON_CLKSET1_PLL1_PS_SHIFT;
	fb1 = ((pll & SYSCON_CLKSET1_PLL1_X1FBD1_MASK) >>
	       SYSCON_CLKSET1_PLL1_X1FBD1_SHIFT) + 1;
	fb2 = ((pll & SYSCON_CLKSET1_PLL1_X2FBD2_MASK) >>
	       SYSCON_CLKSET1_PLL1_X2FBD2_SHIFT) + 1;
	ipd = ((pll & SYSCON_CLKSET1_PLL1_X2IPD_MASK) >>
	       SYSCON_CLKSET1_PLL1_X2IPD_SHIFT) + 1;

	freq = 225 * fb1 * fb2;

	freq = (((freq / ipd) * 65536) + (((freq % ipd) * 65536) / ipd)) >> ps;

	return freq;
}

static int
ep93xx_set_video_div(unsigned long freq)
{
	unsigned long pdiv = 0, div = 0, psel = 0, esel = 0, f, i, j, k, l;
	long err, e;

	err = LONG_MAX;

	for (i = 0; i < 3 && err; i++) {
		if (i == 0)
			f = 14745600 * 2;
		else if (i == 1)
			f = ep93xx_get_pll_frequency(1) * 2;
		else
			f = ep93xx_get_pll_frequency(2) * 2;

		for (j = 4; j <= 6 && err; j++) {
			l = f / (freq * j);
			for (k = l + 1; k >= l && err; k--) {
				if (k < 2)
					break;
				e = (long)(f / (j * k)) - (long)freq;
				if (abs(e) < abs(err)) {
					pdiv = j - 3;
					div = k;
					psel = (i == 2) ? 1 : 0;
					esel = (i == 0) ? 0 : 1;
					err = e;
				}
			}
		}
	}

	if (err == LONG_MAX)
		return -1;

	outl(0xaa, SYSCON_SWLOCK);
	outl(SYSCON_VIDDIV_VENA | (esel ? SYSCON_VIDDIV_ESEL : 0) |
	     (psel ? SYSCON_VIDDIV_PSEL : 0) |
	     (pdiv << SYSCON_VIDDIV_PDIV_SHIFT) |
	     (div << SYSCON_VIDDIV_VDIV_SHIFT), SYSCON_VIDDIV);

	return freq + err;
}

static int
ep93xxfb_configure(struct fb_info *info)
{
	unsigned long htotal, hres, hfp, hsync, hbp;
	unsigned long vtotal, vres, vfp, vsync, vbp;
	unsigned long fb_size, clk, attribs;
	dma_addr_t fb_phys;
	void *fb_log;

	fb_size = modedb[mode].hres * modedb[mode].vres * (bpp / 8);
	fb_log = dma_alloc_writecombine(NULL, fb_size, &fb_phys, GFP_KERNEL);
	if (!fb_log)
		return -ENOMEM;

	vres = modedb[mode].vres;
	vfp = modedb[mode].vfp;
	vsync = modedb[mode].vsync;
	vbp = modedb[mode].vbp;
	vtotal = vres + vfp + vsync + vbp;

	hres = modedb[mode].hres;
	hfp = modedb[mode].hfp;
	hsync = modedb[mode].hsync;
	hbp = modedb[mode].hbp;
	htotal = hres + hfp + hsync + hbp;

	if (modedb[mode].clk_src == CLK_INTERNAL) {
		outl(0xaa, SYSCON_SWLOCK);
		outl(inl(SYSCON_DEVCFG) & ~SYSCON_DEVCFG_EXVC, SYSCON_DEVCFG);
		clk = ep93xx_set_video_div(htotal * vtotal *
					   modedb[mode].refresh);
		if (clk == -1) {
			dma_free_coherent(NULL, fb_size, fb_log, fb_phys);
			return -ENODEV;
		}

		info->var.pixclock = 1000000000 / (clk / 1000);
	} else {
		outl(0xaa, SYSCON_SWLOCK);
		outl(0, SYSCON_VIDDIV);
		outl(0xaa, SYSCON_SWLOCK);
		outl(inl(SYSCON_DEVCFG) | SYSCON_DEVCFG_EXVC, SYSCON_DEVCFG);
		info->var.pixclock = modedb[mode].refresh;
	}

	info->var.xres = hres;
	info->var.yres = vres;
	info->var.xres_virtual = hres;
	info->var.yres_virtual = vres;
	info->var.bits_per_pixel = bpp;
	if (bpp == 8) {
		info->var.red.length = 8;
		info->var.green.length = 8;
		info->var.blue.length = 8;
	} else if ( bpp == 16) {
		info->var.red.offset = 11;
		info->var.red.length = 5;
		info->var.green.offset = 5;
		info->var.green.length = 6;
		info->var.blue.offset = 0;
		info->var.blue.length = 5;
	    
	} else if ( bpp == 24 || bpp == 32 ) {
		info->var.red.offset = 16;
		info->var.red.length = 8;
		info->var.green.offset = 8;
		info->var.green.length = 8;
		info->var.blue.offset = 0;
		info->var.blue.length = 8;
	    
	}
	info->var.activate = FB_ACTIVATE_NOW;
	info->var.height = modedb[mode].height;
	info->var.width = modedb[mode].width;
	info->var.left_margin = hbp;
	info->var.right_margin = hfp;
	info->var.upper_margin = vbp;
	info->var.lower_margin = vfp;
	info->var.hsync_len = hsync;
	info->var.vsync_len = vsync;
	if (modedb[mode].pol_hsync == POL_HIGH)
		info->var.sync |= FB_SYNC_HOR_HIGH_ACT;
	if (modedb[mode].pol_vsync == POL_HIGH)
		info->var.sync |= FB_SYNC_VERT_HIGH_ACT;
	info->var.vmode = FB_VMODE_NONINTERLACED;

	info->fix.smem_start = fb_phys;
	info->fix.smem_len = fb_size;
	info->fix.type = FB_TYPE_PACKED_PIXELS;
	if (bpp == 8)
		info->fix.visual = FB_VISUAL_PSEUDOCOLOR;
	else if (bpp == 16 || bpp == 24 || bpp == 32)
	       	info->fix.visual = FB_VISUAL_TRUECOLOR;

	info->fix.line_length = hres * (bpp / 8);
	info->screen_base = fb_log;

	ep93xxfb_outl(0, VIDEOATTRIBS);

	ep93xxfb_outl(vtotal - 1, VLINESTOTAL);
	ep93xxfb_outl((vtotal - vfp - 1) + ((vtotal - vfp - vsync - 1) << 16),
		      VSYNCSTRTSTOP);
	ep93xxfb_outl((vres - 1) + ((vtotal - 1) << 16), VACTIVESTRTSTOP);
	ep93xxfb_outl((vres - 1) + ((vtotal - 1) << 16), VBLANKSTRTSTOP);
	ep93xxfb_outl((vtotal - 1) + ((vtotal - 1) << 16), VCLKSTRTSTOP);

	ep93xxfb_outl(htotal - 1, HCLKSTOTAL);
	ep93xxfb_outl((htotal - 1) + ((htotal - hsync - 1) << 16),
		      HSYNCSTRTSTOP);
	ep93xxfb_outl((htotal - hsync - hbp - 1) + ((hfp - 1) << 16),
		      HBLANKSTRTSTOP);
	ep93xxfb_outl((htotal - hsync - hbp - 1) + ((hfp - 1) << 16),
		      HACTIVESTRTSTOP);
	ep93xxfb_outl((htotal - 1) + ((htotal - 1) << 16), HCLKSTRTSTOP);

	ep93xxfb_outl(0, LINECARRY);

	outl(fb_phys, VIDSCRNPAGE);
	outl(vres, SCRNLINES);
	outl(((hres * bpp) / 32) - 1, LINELENGTH);
	outl((hres * bpp) / 32, VLINESTEP);

	if (bpp == 8)
		outl(PIXELMODE_S_1PPCMAPPED | PIXELMODE_P_8BPP |
		     PIXELMODE_C_LUT, PIXELMODE);
	else if (bpp == 16)
	    	outl(PIXELMODE_S_1PPCMAPPED | PIXELMODE_P_16BPP |
		     PIXELMODE_C_565, PIXELMODE);
	else if (bpp == 24)
	    	outl(PIXELMODE_S_1PPCMAPPED | PIXELMODE_P_24BPP |
		     PIXELMODE_C_888, PIXELMODE);
	else if (bpp == 32)
	    	outl(PIXELMODE_S_1PPCMAPPED | PIXELMODE_P_32BPP |
		     PIXELMODE_C_888, PIXELMODE);

	SysconSetLocked(SYSCON_DEVCFG,
			inl(SYSCON_DEVCFG) | SYSCON_DEVCFG_RasOnP3);

	attribs = 0;
#ifdef CONFIG_EP93XX_SDCS0
	attribs |= 0 << VIDEOATTRIBS_SDSEL_SHIFT;
#endif
#ifdef CONFIG_EP93XX_SDCS1
	attribs |= 1 << VIDEOATTRIBS_SDSEL_SHIFT;
#endif
#ifdef CONFIG_EP93XX_SDCS2
	attribs |= 2 << VIDEOATTRIBS_SDSEL_SHIFT;
#endif
#ifdef CONFIG_EP93XX_SDCS3
	attribs |= 3 << VIDEOATTRIBS_SDSEL_SHIFT;
#endif
	if (modedb[mode].clk_edge == EDGE_RISING)
		attribs |= VIDEOATTRIBS_INVCLK;
	if (modedb[mode].pol_blank == POL_HIGH)
		attribs |= VIDEOATTRIBS_BLKPOL;
	if (modedb[mode].pol_hsync == POL_HIGH)
		attribs |= VIDEOATTRIBS_HSPOL;
	if (modedb[mode].pol_vsync == POL_HIGH)
		attribs |= VIDEOATTRIBS_VCPOL;
	ep93xxfb_outl(attribs, VIDEOATTRIBS);

	if (modedb[mode].configure)
		(modedb[mode].configure)(info);

	ep93xxfb_blank(0, info);

	return 0;
}

extern int ep93xxfb_setup(char *options);
static int
ep93xxfb_drv_probe(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct resource *res;
	struct fb_info *info;
	int ret;
	char *option = NULL;

#ifndef MODULE
	if (fb_get_options(devicename, &option))
		return -ENODEV;
	ep93xxfb_setup(option);
#endif

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!res)
		return -ENODEV;

	if (!request_mem_region(res->start, res->end - res->start + 1,
				devicename))
		return -EBUSY;

	if (bpp == 8)
		info = framebuffer_alloc(sizeof(u32) * 256, &pdev->dev);
	else if (bpp == 16 || bpp == 24 || bpp == 32)
		info = framebuffer_alloc(sizeof(u32) * 16, &pdev->dev);
	else
		return -EINVAL;

	info->flags = FBINFO_DEFAULT;
	strncpy(info->fix.id, devicename, sizeof(info->fix.id));
	info->fix.mmio_start = res->start;
	info->fix.mmio_len = res->end - res->start + 1;
	info->fbops = &ep93xxfb_ops;
	info->pseudo_palette = info->par;
	info->state = FBINFO_STATE_RUNNING;

	if (fb_alloc_cmap(&info->cmap, 256, 0) < 0) {
		ret = -ENOMEM;
		goto framebuffer;
	}

	if ((ret = ep93xxfb_configure(info)) < 0)
		goto colormap;

	if (register_framebuffer(info) < 0) {
		printk(KERN_ERR "Unable to register EP93xx frame buffer\n");
		ret = -EINVAL;
		goto colormap;
	}

	dev_set_drvdata(dev, info);

	printk(KERN_INFO "fb%d: EP93xx frame buffer at %dx%dx%d\n", info->node,
	       info->var.xres, info->var.yres, info->var.bits_per_pixel);

	return 0;

colormap:
	fb_dealloc_cmap(&info->cmap);
framebuffer:
	framebuffer_release(info);

	return ret;
}

static int __devexit
ep93xxfb_drv_remove(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct fb_info *info = dev_get_drvdata(dev);
	struct resource *res;

	dev_set_drvdata(dev, NULL);

	unregister_framebuffer(info);

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	release_mem_region(res->start, res->end - res->start + 1);

	fb_dealloc_cmap(&info->cmap);
	framebuffer_release(info);

	return 0;
}

static struct device_driver ep93xxfb_driver = {
	.name		= devicename,
	.bus		= &platform_bus_type,
	.probe		= ep93xxfb_drv_probe,
	.remove		= __devexit_p(ep93xxfb_drv_remove),
};

int __init
ep93xxfb_init(void)
{
#ifdef MODULE
	int i;

	if (modestr) {
		for (i = 0; i < (sizeof(modedb) / sizeof(modedb[0])); i++)
			if (strcmp(modestr, modedb[i].name) == 0)
				break;
		if (i != (sizeof(modedb) / sizeof(modedb[0])))
			mode = i;
		else
			printk(KERN_ERR devicename ": unknown mode: %s\n",
			       modestr);
	}
#endif

	return driver_register(&ep93xxfb_driver);
}

static void __exit
ep93xxfb_cleanup(void)
{
	return driver_unregister(&ep93xxfb_driver);
}

int
ep93xxfb_setup(char *options)
{
	char *opt;
	int i;

	if (!options || !*options)
		return 0;

	while ((opt = strsep(&options, ",")) != NULL) {
		if (!*opt)
			continue;
		for (i = 0; i < (sizeof(modedb) / sizeof(modedb[0])); i++)
			if (strcmp(opt, modedb[i].name) == 0)
				break;
		if (i != (sizeof(modedb) / sizeof(modedb[0]))) {
			mode = i;
			continue;
		}
		if (strcmp(opt, "8bpp") == 0) {
			bpp = 8;
			continue;
		} else if (strcmp(opt, "16bpp") == 0) {
			bpp = 16;
			continue;
		} else if (strcmp(opt, "24bpp") == 0) {
			bpp = 24;
			continue;
		} else if (strcmp(opt, "32bpp") == 0) {
			bpp = 32;
			continue;
		}

		printk (KERN_ERR devicename ": unknown parameter: %s\n", opt);
	}

	return 0;
}

module_init(ep93xxfb_init);
module_exit(ep93xxfb_cleanup);

MODULE_PARM(modestr, "s");
MODULE_PARM(bpp, "i");

MODULE_AUTHOR("Cirrus Logic");
MODULE_DESCRIPTION("EP93xx frame buffer driver");
MODULE_LICENSE("GPL");
